package KL::App;

use strict;
use warnings;

use File::Spec;
use File::Temp;
use File::Basename;
use Data::Dumper;

use KL::AppBase;
use KL::IniFile;
use KL::LmsUtils;

my $APP_ID = 'klmsui';

use vars qw(@ISA);
@ISA = qw(KL::AppBase);

my $appname = "Kaspersky Security 8.0 for Linux Mail Server";

### Text strings BEGIN ###
my $HEADER_APACHE_MODULE = <<EOT;
Setting up the Apache module
EOT

my $MSG_PASS_NOT_SET = "Web-based administration is not available because password for Web Management Console was not set. You can set the password running %1 with option --set-web-admin-password.";

my $MSG_APACHE_FOUND = <<EOT;
Installer found the following Apache server on your computer:

Apache server daemon executable: <B>%1</B>
Apache server configuration dir: <B>%2</B>
Apache server service name (or rc.d script name): <B>%3</B>

Please enter 'Y' to confirm that you want to integrate with this Apache server. Enter 'N' if Apache server has been detected incorrectly, or if you do not want to use it
EOT
my $MSG_WMCONSOLE_PASS_CONFIRM  = "Would you like to set password for Kaspersky Web Management Console?";
my $MSG_APACHE_CONFIRM_CUSTOM = <<EOT;
Setup was unable to detect an existing installation of an Apache server on your computer. Either it is not installed or has been installed in an unknown location.
If it has been installed, and you know the installation details, enter 'Y'. Otherwise enter 'N' (configuration will be canceled)
EOT

my $MSG_WMCONSOLE_PASS_WARNING = <<EOT;
Web Administrator password was not saved in the file with saved answers. You have two options to specify password during application configuration cloning:

<LI>- specify Web Administrator password by executing the <BR><B>%1 --set-web-admin-password</B> <BR>command on each server after completing the application setup in interactive mode;

<LI>- specify Web Administrator password as <B>WEB_PASSWORD</B> parameter value in the file with saved answers before launching the application\'s setup script in automatic mode. <BR>Please, keep in mind that keeping password in answers file <U>is insecure</U>.
<BR>
EOT

#---
my $MSG_APACHE_ASK_HTTPD_PATH = <<EOT;

Please enter the full path (including the file name) to your httpd binary (or <B>enter 'cancel' to interrupt</B> the Apache server configuration step)
EOT

my $MSG_APACHE_ASK_INIT_D = <<EOT;

Please enter the Apache service name (e.g. httpd, apache2) or full path to Apache rc.d script (or <B>enter 'cancel' to interrupt</B> the Apache server configuration step)
EOT

my $MSG_APACHE_ASK_CONF_D_PATH = <<EOT;

Please enter the full path to your Apache conf.d directory (or <B>enter 'cancel' to interrupt</B> the Apache server configuration step)
EOT

my $MSG_APACHE_RECONFIG_HINT = <<EOT;
You can configure integration to Apache server later: run the initial configuration script again by executing
<B>%1</B> 
EOT

my $MSG_APACHE_INTEGRATION_CANCELED = <<EOT;
The integration to Apache server was not setup because the path to Apache binaries was not specified. You can run the initial configuration script again by executing
<B>%1</B>
EOT

my $MSG_DIR_SECURITY_WARNING = <<EOT;
When using a directory-based Apache virtual host, Kaspersky Security will use connection settings specified in the Apache configuration file. The default connection type is a non-secure HTTP. You can manually set up the Apache virtual host to use the encrypted SSL connection.
EOT

my $MSG_NAME_VHOST_NOTICE = <<EOT;
Please check if you have proper NameVirtualHost directive in main Apache configuration. Otherwise name-based virtual host for KLMS UI may not work as expected.
EOT

###
my $MSG_APACHE_ASK_VHOST_TYPE = <<EOT;
Please select configuration type. Enter:
 'name' - for name-based virtual host;
 'port' - for port-based virtual host;
 'dir' - for directory-based configuration.

EOT

my $MSG_APACHE_BAD_VHOST_TYPE = <<EOT;
'%1' is not a valid configuration type. Please enter 'port', 'name' or 'dir'.
EOT

my $MSG_APACHE_ASK_VHOST_PORT = <<EOT;
Please enter port number for virtual host or press Enter to use default
EOT

my $MSG_APACHE_BAD_VHOST_PORT = <<EOT;
'%1' is not a valid port number
EOT

my $MSG_APACHE_ASK_VHOST_HOST = <<EOT;
Please enter host name for virtual host
EOT

my $MSG_APACHE_ASK_VHOST_DIR = <<EOT;
Please enter sub-directory for KLMS UI
EOT

my $MSG_APACHE_BAD_VHOST_HOST = <<EOT;
'%1' is not a valid host name
EOT

my $MSG_APACHE_BAD_VHOST_DIR = <<EOT;
'%1' is not a valid sub-directory
EOT

my $MSG_APACHE_ASK_CERT_TYPE = <<EOT;
Please select which certificate will be used for this virtual host.
Enter 'new' to generate a new self-signed certificate, or enter 'file' to specify key and certificate files manually
EOT

my $MSG_APACHE_ASK_CERT_TYPE_W_KEEP = <<EOT;
Please select which certificate will be used for this virtual host.
Enter 'keep' to keep existing certificate, 'new' to generate a new self-signed certificate, or enter 'file' to specify key and certificate files manually
EOT

my $MSG_APACHE_BAD_CERT_TYPE = <<EOT;
'%1' is not a valid certificate type. Please enter 'new' or 'file'
EOT

my $MSG_APACHE_BAD_CERT_TYPE_W_KEEP = <<EOT;
'%1' is not a valid certificate type. Please enter 'keep', 'new' or 'file'
EOT

my $MSG_APACHE_ASK_CERT_KEY = <<EOT;
Please enter path to PEM-encoded Private Key file
EOT

my $MSG_APACHE_BAD_CERT_KEY = <<EOT;
'%1' is not a valid path to certificate key file
EOT

my $MSG_APACHE_ASK_CERT_CRT = <<EOT;
Please enter path to PEM-encoded Certificate file
EOT

my $MSG_APACHE_BAD_CERT_CRT = <<EOT;
'%1' is not a valid path to certificate file
EOT

my $HEADER_APACHE_DONE = <<EOT;
Configure finished
EOT

my $MSG_APACHE_CONFIGURED = <<EOT;
Apache Web Server is successfully configured.
In your browser, navigate to <U>%2</U> URL for access to <B>%1</B>
EOT

my $MSG_APACHE_MODULES_NOTE = <<EOT;
In order to make KLMS web-UI work properly, please check that following Apache modules are installed and enabled:
- mod_dir
- mod_expires
- mod_include
- mod_ssl (for name-based or port-based configuration)

EOT

###

my $MSG_APACHE_ASK_UI_HOST = <<EOT;
Please enter $appname UI host and port in <hostname>:<port> format
EOT

my $MSG_APACHE_BAD_UI_HOST = <<EOT;
'%1' is not a valid $appname UI. Please enter in <hostname>:<port> format
EOT

###

my $MSG_CREATE_SSL_CERT = <<EOT;
<U>Generate SSL Certficate</U>
Domain: %1
Lifetime: %2 days
EOT

my $MSG_INSTALL_SSL_CERT = <<EOT;
<U>Install SSL Certficate</U>
EOT

my $MSG_KEEP_SSL_CERT = <<EOT;
<U>Keep Existing SSL Certficate</U>
EOT

my $MSG_INSTALL_VHOST_CONF = <<EOT;
<U>Install KLMS UI Apache config file</U>
EOT

my $MSG_GENERATE_KEY = "Generate Certificate Key File";
my $MSG_GENERATE_CSR = "Generate Certificate File";
my $MSG_SUCCESS = 'success';
my $MSG_FAILED = 'failed';

###
my $HEADER_UPDATE_APACHE = <<EOT;
Update Apache configuration
EOT

my $MSG_CONFIG_SAVED = <<EOT;
Current config file is saved as %1
EOT

my $MSG_INSTALL_FILE = <<EOT;
Installing file %1
EOT

my $MSG_CANNOT_EXECUTE = <<EOT;
Unable to execute %1: %2
EOT

my $MSG_CANNOT_FIND_SERVICE = <<EOT;
Unable to find service or rc.d script %1
EOT

my $MSG_CANNOT_EXECUTE_COMMAND = <<EOT;
Could not execute the command %1 %2: %3
EOT

my $MSG_NOFILE = "No such file";
my $MSG_NODIR = "The %1 directory does not exist";

my $MSG_CANT_WRITE = <<EOT;
Could not write %1: %2
EOT

my $MSG_CANT_READ = <<EOT;
Could not read %1: %2
EOT

my $MSG_DIR_NOT_FOUND = <<EOT;
No such directory: <B>%1</B>
EOT
my $MSG_DIR_EMPTY_PATH = <<EOT;
Empty path not allowed
EOT

my $MSG_SELINUX = <<EOT;
Working with enabled SeLinux is not guaranteed!
EOT
###

my $MSG_CMD_HELP_APACHE = "-A, --apache\tSet up the Apache server integration";

###
my $MSG_AI_HELP_EULA_AGREED = "=yes\tLicense agreement flag, must be 'yes'";
my $MSG_AI_HELP_APACHE_CONF_D = "=<path>|auto\tPath to Apache conf.d directory";
my $MSG_AI_HELP_APACHE_INIT_D = "=<path>|auto\tApache service name or rc.d script path";
my $MSG_AI_HELP_UI_HOST = "=<host:port>\t$appname host and port";
my $MSG_AI_HELP_VHOST_TYPE = "=<port|name|dir>\tVirtual host template type";
my $MSG_AI_HELP_VHOST_PORT = "=<port>\tVirtual host port number (for port template)";
my $MSG_AI_HELP_VHOST_HOST = "=<hostname>|auto\tVirtual host server name";
my $MSG_AI_HELP_VHOST_DIR = "=<url_subdir>|auto\tKLMS URL subdirectory";
my $MSG_AI_HELP_CERT_TYPE = "=<new|file|keep>\tCertificate installation method";
my $MSG_AI_HELP_CERT_CRT = "=<path>\tPath to certificate file (for method 'file')";
my $MSG_AI_HELP_CERT_KEY = "=<path>\tPath to certificate private key file (for method 'file')";
my $MSG_AI_IGNORE_APACHE_ARCH = "=yes|no\tIgnore error in case when apache architecture can't be determined (force integration)";
### Text strings END ###

sub new
{
    my $class = shift;
    my $self = $class->SUPER::new($APP_ID);

    unshift @{$self->{'COMMANDS'}}, {
        'id' => 'apache',
        'spec' => 'A|apache',
        'sub' => 'do_setup_apache_module',
        'default' => undef,
        'descr' => $MSG_CMD_HELP_APACHE,
    },
    {
        'id' => 'postupgrade',
        'spec' => 'postupgrade',
        'sub' => 'do_postupgrade',
    };

    $self->{'AUTOINSTALL'} = [
        {
            'id'        => 'WEB_EULA_AGREED',
            'list'      => [ 'yes' ],
            'required'  => 1,
            'descr'     => $MSG_AI_HELP_EULA_AGREED,
        },

        {
            'id'        => 'IGNORE_APACHE_ARCH',
            'list'      => [ 'yes', 'no' ],
            'required'  => 0,
            'default'   => 'no',
            'descr'     => $MSG_AI_IGNORE_APACHE_ARCH,
        },
        {
            'id'        => 'APACHE_CONF_D',
            'type'      => 'dir',
            'required'  => 1,
            'descr'     => $MSG_AI_HELP_APACHE_CONF_D,
        },

        {
            'id'        => 'APACHE_INIT_D',
            'type'      => 'service',
            'required'  => 1,
            'descr'     => $MSG_AI_HELP_APACHE_INIT_D,
        },

        {
            'id'        => 'UI_HOST',
            'type'      => 'hostandport',
            'required'  => 1,
            'descr'     => $MSG_AI_HELP_UI_HOST,
        },
        {
            'id' => 'VHOST_TYPE',
            'list' => ['host','port', 'dir'],
            'required' => 1,
            'descr'     => $MSG_AI_HELP_VHOST_TYPE,
        },
        {
            'id'        => 'VHOST_PORT',
            'type'      => 'port',
            'required'  => 0,
            'descr'     => $MSG_AI_HELP_VHOST_PORT,
        },
        {
            'id'        => 'VHOST_DIR',
            'type'      => 'url_subdir',
            'required'  => 0,
            'descr'     => $MSG_AI_HELP_VHOST_DIR,
        },

        {
            'id'        => 'VHOST_HOST',
            'list'      => [ 'auto' ],
            'required'  => 0,
            'descr'     => $MSG_AI_HELP_VHOST_HOST,
        },

        {
            'id'        => 'CERT_TYPE',
            'list'      => [ 'new', 'file', 'keep' ],
            'required'  => 1,
            'descr'     => $MSG_AI_HELP_CERT_TYPE,
        },

        {
            'id'        => 'CERT_CRT',
            'type'      => 'file',
            'required'  => 0,
            'descr'     => $MSG_AI_HELP_CERT_CRT,
        },

        {
            'id'        => 'CERT_KEY',
            'type'      => 'file',
            'required'  => 0,
            'descr'     => $MSG_AI_HELP_CERT_KEY,
        },
    ];

    $self->load_settings();

    my $bin = sprintf('%s/%s', $self->get_path('bins'),
        $self->appinfo('product', 'name'));

    $self->{'SETUP'} = "${bin}-setup.pl";
    $self->{'UPGRADE'} = "${bin}-upgrade.pl";

    #$self->{'APACHE_RELOAD_CMD'} = 'try-restart';
    $self->{'APACHE_RELOAD_CMD'} = 'restart';
    #$self->{'APACHE_RELOAD_CMD'} = 'reload';

    $self->{'EULA_L10N'} = 1;
    $self->{'_LMS_L10N'} = {};

    return $self;
}

###

sub _is_valid_ip_address
{
    my $value = shift;

    return 1 if $value =~ /^(([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\.){3}([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])$/;
    return 0;
}

sub _is_valid_hostname
{
    my $value = shift;

    return 1 if $value =~ /^(([a-zA-Z]|[a-zA-Z][a-zA-Z0-9\-]*[a-zA-Z0-9])\.)*([A-Za-z]|[A-Za-z][A-Za-z0-9\-]*[A-Za-z0-9])$/;
    return 0;
}

sub _is_valid_url_subdir
{
    my $value = shift;
    # filter multiple slashes
    return 0 if $value =~ /\/\//;
    # single slash is not allowed too.
    return 0 if $value eq "/";

    return 1 if $value =~ /^[\w+\/]+$/;
    return 0;
}


sub get_apache_conf_d
{
    my $httpd_path = shift;

    my ($ret, $out) = get_apache_compile_options($httpd_path);

    if ($ret != 0) {
        KL::Utils::error("Cannot get Apache configuration options (command 'apachectl -V' failed). See output below:\n$out");
        return undef;
    }

    my $httpd_root = "";
    if ($out =~ /HTTPD_ROOT="([^"]*)"/) {
        $httpd_root = "$1";
        KL::Utils::debug("Apache HTTPD_ROOT is $httpd_root");
    } else {
        KL::Utils::error("Cannot parse output of 'apachectl -V' command: cannot find HTTPD_ROOT parameter. See this output below:\n$out");
        return undef;
    }

    my $server_config_file = "";
    if ($out =~ /SERVER_CONFIG_FILE="([^"]+)"/) {
        $server_config_file = "$1";
        KL::Utils::debug("Apache SERVER_CONFIG_FILE is\n$server_config_file");
    } else {
        KL::Utils::error("Cannot parse output of 'apachectl -V' command: cannot find SERVER_CONFIG_FILE parameter. See this output below:\n$out");
        return undef;
    }

    my $main_apache_conf = $server_config_file;
    if (!KL::LmsUtils::is_absolute_path( $server_config_file)) {
        $main_apache_conf = "$httpd_root/$server_config_file";
    }

    $main_apache_conf = KL::LmsUtils::trim("$main_apache_conf");
    KL::Utils::debug("Main Apache config is:\n$main_apache_conf");

    if (! -f $main_apache_conf) {
        KL::Utils::error("Apache configuration file \'$main_apache_conf\' not found. ");
        return undef;
    }

    my $conf_dir = File::Basename::dirname( $main_apache_conf);

    my @conf_d_variants = ("$conf_dir/conf-enabled", "$conf_dir/conf.d", "$conf_dir/Includes");

    if (KL::LmsUtils::is_not_empty( $httpd_root)) {
        push @conf_d_variants, "$httpd_root/conf-enabled", "$httpd_root/conf.d", "$httpd_root/Includes";
    }

    my $conf_d = "";
    for my $dir (@conf_d_variants) {
        if (-d $dir) {
            $conf_d = $dir;
            last;
        }
    }

    if ($conf_d eq "") {
        KL::Utils::error("Apache conf.d directory not found");
        return undef;
    }

    KL::Utils::debug("Apache conf.d directory is $conf_d");
    return $conf_d;
}

sub find_apache
{
    my $result = {
        'bin' => "",
        'service_control' => "",
        'conf.d' => ""
    };

    my $apacheSrv = KL::LmsUtils::find_service('httpd', 'httpd2', 'apache2', 'apache22', 'apache24');
    if (!$apacheSrv) {
        KL::Utils::error("Apache service not found");
        return (0, $result);
    }

    my $service_name = $apacheSrv->{'name'};
    my $apache_rc_d = $apacheSrv->{'rc'};
    KL::Utils::output("Apache service name: $service_name");

    my $apache_service_control = $service_name;

    if (KL::LmsUtils::is_not_empty( $apache_rc_d)) {
        KL::Utils::debug("Using rc.d script to control Apache. ");
        $apache_service_control = $apache_rc_d;
    }

    $result->{'service_control'} = $apache_service_control;

    my $apache_bin = get_apache_binary();

    if (KL::LmsUtils::is_empty( $apache_bin)) {
        KL::Utils::error("Cannot get Apache binary. ");
        return (0, $result);
    }

    $result->{'bin'} = $apache_bin;

    my $conf_d = get_apache_conf_d($apache_bin);

    if (KL::LmsUtils::is_empty( $conf_d)) {
        KL::Utils::error("Cannot get Apache conf.d directory. ");
        return (0, $result);
    }

    $result->{'conf.d'} = $conf_d;

    return (1, $result);
}

###

sub _ask_apache_paths
{
    my $self = shift;

    my $setup_cmd = $self->{'SETUP'} . " --apache";

    my ($res, $apache) = find_apache();
    KL::Utils::debug("Apache configuration: " . Dumper($apache) );

    my $httpd = $apache->{'bin'} || "";
    my $conf_d = $apache->{'conf.d'} || "";
    my $service_control = $apache->{'service_control'} || "";

    if ($res == 1) {
        my @args = (
            $httpd,
            $conf_d,
            $service_control
        );

        my $confirmed = KL::Utils::ask_boolean ('yes', $MSG_APACHE_FOUND, @args);

        if ($confirmed) {
            return $apache;
        }
    } else {
        if ( KL::Utils::ask_boolean ('no', $MSG_APACHE_CONFIRM_CUSTOM) == 0 ) {
            KL::Utils::output( $MSG_APACHE_RECONFIG_HINT, $setup_cmd);
            return;
        }
    }

    #
    # Auto-detection failed or not confirmed
    #

    # (1) Ask user for httpd binary
    my $default_httpd = $httpd;

    while ( 1 ) {
        $httpd = KL::Utils::ask_question ( $default_httpd, $MSG_APACHE_ASK_HTTPD_PATH);

        if ($httpd =~ /^s*cancel\s*$/i) {
            KL::Utils::output ($MSG_APACHE_INTEGRATION_CANCELED, $setup_cmd);
            return;
        }

        # exists?
        if (! -x $httpd) {
            KL::Utils::output ($MSG_CANNOT_EXECUTE, $httpd, $MSG_NOFILE);
            next;
        }

        last;
    }

    # (2) Ask user for Apache conf.d directory.
    my $default_conf_d = $conf_d;

    while ( 1 ) {
        $conf_d = KL::Utils::ask_question ( $default_conf_d, $MSG_APACHE_ASK_CONF_D_PATH);

        if ($conf_d =~ /^s*cancel\s*$/i) {
            KL::Utils::output( $MSG_APACHE_INTEGRATION_CANCELED, $setup_cmd);
            return;
        }

        if (-d $conf_d) {
            last;
        }
        if ( $conf_d =~/^\s*$/ ) {
            KL::Utils::output( $MSG_DIR_EMPTY_PATH);
            next;
        }
        KL::Utils::output( $MSG_DIR_NOT_FOUND, $conf_d);
    }

    # (3) Ask user for path to Apache's service name or rc.d script

    my $default_service_control = $service_control;
    while ( 1 ) {
        $service_control = KL::Utils::ask_question ( $default_service_control, $MSG_APACHE_ASK_INIT_D);

        if ($service_control =~ /^s*cancel\s*$/i) {
            KL::Utils::output( $MSG_APACHE_INTEGRATION_CANCELED, $setup_cmd);
            return;
        }
        # check if this service or rc.script really exists

        if (!KL::LmsUtils::find_service( $service_control)) {
            KL::Utils::output( $MSG_CANNOT_FIND_SERVICE, $service_control);
            next;
        }

        last;
    }

    $apache = {
        'bin'  => $httpd,
        'conf.d' => $conf_d,
        'service_control' => $service_control
    };

    return $apache;
}

sub _openssl_nout
{
    my $openssl = '/usr/bin/openssl';

    my ($ret, $out) = KL::Utils::execute($openssl, @_);
    KL::Utils::warning($MSG_CANNOT_EXECUTE_COMMAND, $openssl, join(' ', @_), $!) if $ret == -128;

    # KL::Utils::debug("_openssl_nout( $openssl " . join(' ', @_) . " ) = $ret '$out'");

    return ($ret, $out);
}

sub _openssl
{
    my @args = @_;
    foreach (@args) {
        $_ = KL::Utils::shell_escape($_);
    }

    my $openssl = '/usr/bin/openssl';

    KL::Utils::debug("_openssl $openssl " . join(' ', @args));

    system($openssl, @args);

    my $ret = $?;
    if ($ret == -1) {
        KL::Utils::warning($MSG_CANNOT_EXECUTE_COMMAND, $openssl, join(' ', @args), $!);
        return -128;
    } elsif ($ret & 127) {
        $ret &= 127;
    } else {
        $ret >>= 8;
    }
    return $ret;
}

sub _create_ssl_cert
{
    my $self = shift;

    my $cert = shift;

    KL::Utils::output($MSG_CREATE_SSL_CERT,
        $cert->{'domain'}, $cert->{'days'}, $cert->{'key'}, $cert->{'crt'});

    my $key = $cert->{'key'};
    my $crt = $cert->{'crt'};
    my $length = defined $cert->{'length'} ? defined $cert->{'length'} : 1024;
    my $subj = '/CN=' . $cert->{'domain'};

    my ($rc, $out);

    print "\n";

    unlink $key if -f $key;
    unlink $crt if -f $crt;

    print "$MSG_GENERATE_KEY $key: ";
    # generate RSA key file
    ($rc, $out) = _openssl_nout('genrsa', '-des3', '-out', $key, '-passout', 'pass:1111', $length);
    if (($rc != 0) or (! -f $key)) {
        print "$MSG_FAILED: $out\n";
        return 0;
    }
    ### remove passphrase from key
    ($rc, $out) = _openssl_nout('rsa', '-in', $key, '-out', $key, '-passin', 'pass:1111');
    if (($rc != 0) or (! -f $key)) {
        print "$MSG_FAILED: $out\n";
        return 0;
    }
    print $MSG_SUCCESS, "\n";

    print "$MSG_GENERATE_CSR $crt: ";
    ($rc, $out) = _openssl_nout('req', '-batch', '-new', '-x509',
        '-days', $cert->{'days'},
        '-key', $key,
        '-out', $crt,
        '-subj', $subj,
    );
    if (($rc != 0) or (! -f $crt)) {
        print "$MSG_FAILED: $out\n";
        return 0;
    }
    print $MSG_SUCCESS, "\n";

    return 1;
}

sub _get_hostname
{
    my $default = shift;

    my ($ret, $hostname) = KL::Utils::execute('/bin/hostname', '--fqdn');
    chomp $hostname;
    return lc($hostname) if $ret == 0;

    return $default;
}
sub write_file
{
    my $filename = shift;
    my $lines = shift;

    open F, ">$filename"
        or return;

    my $result = print F $lines;
    close F;

    return $result;
}

sub get_apache_binary
{
    my $httpd_path = shift;

    if (KL::LmsUtils::is_empty( $httpd_path)) {
        # TODO: possibly, replace it with searching of 'real' httpd binary.
        # actually, it does not matter what binary we use - apachectl or httpd. Option -V is recoginzed by both binaries.
            $httpd_path = find_apachectl();
    }
    return $httpd_path;
}

sub find_apachectl
{
    my $bin = KL::Utils::which("apachectl");
    if (KL::LmsUtils::is_empty( $bin)) {
        $bin = KL::Utils::which("apache2ctl");
        if (KL::LmsUtils::is_empty( $bin)) {
            KL::Utils::error("Cannot find apachectl binary");
        }
    }
    return $bin;
}

sub get_apache_compile_options
{
    my $httpd_path = shift;

    my $apache_bin = get_apache_binary($httpd_path);

    if (KL::LmsUtils::is_empty( $apache_bin)) {
        KL::Utils::error("Apache binary not found. ");
        return (2, "Apache binary not found");
    }
    KL::Utils::debug("Apache binary in get_apache_compile_options(): $apache_bin");

    # to avoid problems with localized output
    my ($ret, $out) = KL::Utils::execute("env", "LANG=C", $apache_bin, "-V");
    KL::Utils::debug("'apachectl -V' output:\n$out\n");
    return ($ret, $out);
}

sub get_apache_apr_version
{
    my $httpd_path = shift;

    my ($ret, $out) = get_apache_compile_options($httpd_path);

    if ($ret != 0) {
        KL::Utils::error("Cannot get Apache compile options (apachectl -V failed). ");
        return undef;
    }
    if ($out =~ /Compiled using:\s+APR\s+1\.(\d+)/) {
        my $apr_version = "$1";
        KL::Utils::debug("Apache APR version is $apr_version");
        if ($apr_version lt "3") {
            KL::Utils::warning("Apache server uses APR library 1.$apr_version which is not supported. Trying to use module for APR 1.3.x");
            $apr_version = "3";
        }
        return $apr_version;
    }

    KL::Utils::warning("Apache server uses unknown version of APR library. Trying to use module for oldest available version.");
    return "3";
}

sub get_apache_server_version
{
    my $httpd_path = shift;

    my ($ret, $out) = get_apache_compile_options($httpd_path);

    if ($ret != 0) {
        KL::Utils::error("Cannot get Apache compile options (apachectl -V failed). ");
        return undef;
    }

    if ($out =~ /Server version:\s+Apache\/(\d+\.\d+)/) {
        my $apache_version = "$1";
        KL::Utils::debug("Apache server version is $apache_version");
        if ($apache_version le "2.3") {
            $apache_version = "2.2"; # apache versions below 2.3 uses old API. Only in Apache 2.4 API had changed.
        }
        return $apache_version;
    }

    KL::Utils::error("Cannot parse output of 'apachectl -V' command. See this output below:\n$out");
    return undef;
}

sub _setup_apache
{
    my $self = shift;

    KL::Utils::header( $HEADER_UPDATE_APACHE);

    my $opts = shift;

    KL::Utils::debug('_setup_apache(' . Dumper($opts) );

    my $apache = $opts->{'apache'};
    my $httpd_path = $apache->{'bin'};

    my $name = $self->appinfo('product', 'name');

    my $conf = sprintf('%s/%s.conf', $apache->{'conf.d'}, $name);
    my $cert_key = sprintf('%s/.%s.key', $apache->{'conf.d'}, $name);
    my $cert_crt = sprintf('%s/.%s.crt', $apache->{'conf.d'}, $name);

    # install cert files
    if ($opts->{'cert_type'} eq 'keep') {
        KL::Utils::output($MSG_KEEP_SSL_CERT, $cert_key, $cert_crt);
    } if ($opts->{'cert_type'} eq 'new') {
        # copy certificate files
        my $cert_days = 365;

        my $cert = {
            'domain' => $opts->{'vhost_host'},
            'days'   => $cert_days,
            'key'    => $cert_key,
            'crt'   =>  $cert_crt,
        };

        my $res = $self->_create_ssl_cert($cert);
        if (!$res) {
            $self->_ai_error(3);
        }
    } elsif ($opts->{'cert_type'} eq 'file') {
        # copy certificate files
        KL::Utils::output($MSG_INSTALL_SSL_CERT);

        KL::Utils::output($MSG_INSTALL_FILE, $cert_key);
        KL::Utils::copy_file($opts->{'cert_key'}, $cert_key);

        KL::Utils::output($MSG_INSTALL_FILE, $cert_crt);
        KL::Utils::copy_file($opts->{'cert_crt'}, $cert_crt);
    }
    print "\n";

    # install config from template
    KL::Utils::output( $MSG_INSTALL_VHOST_CONF);

    my $tpl_file = sprintf('%s/configs/%s.conf.template', $self->get_path('share'), $opts->{'vhost_type'});

    my $apache_version = get_apache_server_version($httpd_path);

    # enable 2.2.x compatible parser for newer apache versions
    my $ssi_legacy_parser_directive = "";
    if ($apache_version ge "2.4") {
        $ssi_legacy_parser_directive = "SSILegacyExprParser on";
    }

    # add "Require all granted" directive for Apache 2.4, or old-style 'Allow from all'
    my $access_directive = "";
    if ($apache_version ge "2.4") {
        $access_directive = "Require all granted";
    } else {
		$access_directive = "Order allow,deny\n\tAllow from all";
	}

    my $apr_version = get_apache_apr_version($httpd_path);
    if (!defined($apr_version)) {
        KL::Utils::fatal(1, "Cannot determine which APR library version is used in Apache. ");
    }

    my %tpl_vars = (
        'EULA_AGREED' => $self->settings('WEB_EULA_AGREED'),

        'DOMAIN' => $opts->{'vhost_host'},
        'PORT'   => $opts->{'vhost_port'},
        'URL_SUBDIR'   => $opts->{'vhost_dir'},
        'UI_HOST' => $opts->{'ui_host'},

        'LIBEXEC' => $self->get_path('libexec'),
        'APACHE_SERVER_ROOT' => $self->get_path('share') . '/htdocs',

        'SSL_CERTIFICATE_FILE' => $cert_crt,
        'SSL_CERTIFICATE_KEY_FILE' => $cert_key,

        'APR_VER' => $apr_version,
        'APACHE_VER' => $apache_version,
        'SSI_LEGACY_PARSER' => $ssi_legacy_parser_directive,
        'ACCESS_DIRECTIVE' => $access_directive
    );

    KL::Utils::output( $MSG_INSTALL_FILE, $conf);
    my $tpl = KL::Utils::read_file( $tpl_file);
    if (not defined $tpl) {
        KL::Utils::error( $MSG_CANT_READ, $tpl_file, $!);
        return 0;
    }

    foreach my $var (keys %tpl_vars) {
        #my $value = quotemeta($params{$var});
        my $value = $tpl_vars{$var};
        if (!defined($value)) {
            KL::Utils::error("Cannot perform replacement in config template: parameter \'$var\' has undefined value. ");
            $value = "";
        }
        $tpl =~ s/%$var%/$value/g;
    }

    write_file($conf, $tpl);

    $self->settings('APACHE_CONF', $conf);
    $self->settings('APACHE_INIT_D', $apache->{'service_control'});
    $self->save_settings();
    if ( -x "/usr/sbin/selinuxenabled"  && -x "/usr/bin/chcon" ) {
        `/usr/sbin/selinuxenabled`;
        if ( $? == 0 ) {
            KL::Utils::warning( $MSG_SELINUX);
        }
    }
    print "\n";
    KL::LmsUtils::service( $apache->{'service_control'}, $self->{'APACHE_RELOAD_CMD'});
    if ($? != 0) {
        KL::Utils::fatal(1, "Apache restart failed. Look Apache logs for more details. $MSG_APACHE_MODULES_NOTE");
    }


    ###
    KL::Utils::header($HEADER_APACHE_DONE);

    my $url;
    my $vHostType = $opts->{'vhost_type'};
    my $vHostHost = $opts->{'vhost_host'};

    if ($vHostType eq 'port') {
        my $vHostPort = $opts->{'vhost_port'};
        $url = "https://$vHostHost:$vHostPort/";
    } elsif ($vHostType eq 'name') {
        $url = "https://$vHostHost/";
    } elsif ($vHostType eq 'dir') {
        my $vHostDir = $opts->{'vhost_dir'};
        $url = "http://$vHostHost/$vHostDir/";
    } else {
        KL::Utils::fatal(1, "Unsupported vhost type: %1", $opts->{'vhost_type'});
    }

    # execute our 'micro-curl' to save link in core.
    if (!$self->store_webui_link($url))
    {
        KL::Utils::fatal(1, "Cannot store Web Console link. ");
    }

    KL::Utils::output($MSG_APACHE_CONFIGURED, $self->appinfo('product', 'title'), $url);
    KL::Utils::output($MSG_APACHE_MODULES_NOTE);
    print "\n";
}

sub store_webui_link
{
    my $self = shift;
    my $url = shift;
    KL::LmsUtils::assert_not_empty( $url, "url");

    my $libexecPath = $self->get_path('libexec');
    my $curlBinary = "$libexecPath/web_url_tool";
    my ($result, $output) = KL::Utils::execute( $curlBinary, "$url"."cgi-bin/klwi?action=storeWebUiUrl");
    if ($result != 0) {
        #KL::Utils::error("Cannot save WebUI URL. ");
        KL::Utils::debug("Warning: cannot save WebUI URL. Command output:\n$output");
		# we don't consider this as error because in dir-based config this link may not work due to non-default apache port, etc.
        #return 0;
    }
    return 1;
}

sub get_klms_ctrl
{
    my $result;

    my $bin = KL::Utils::is_freebsd() ? '/usr/local/bin' : '/opt/kaspersky/klms/bin';

    if ( -x "$bin/klms-control" ) {
        return  "$bin/klms-control";
    }else{
        KL::Utils::fatal(1,"Can't find klms-control");
    }
    return undef;
}

sub get_klms_facade_settings
{
    my $self = shift;
    my $klms_ctrl = get_klms_ctrl();
    my ($ret,$out) = KL::Utils::execute($klms_ctrl, '--get-settings',4);
    if ($ret != 0) {
        KL::Utils::fatal(1,"Can't get_app_settings for klms");
    }
    return $out;
}

sub set_web_password
{
    my $self = shift;

    my $klms_ctrl = get_klms_ctrl();

    my ($ret, $out) = KL::Utils::execute($klms_ctrl, '--is-admin-password-set');

    if ($ret != 0) {
        my $answer = KL::Utils::ask_boolean ('no', $MSG_WMCONSOLE_PASS_CONFIRM);
        my $result;
        while (1) {
            if ($answer  == 0) {
                KL::Utils::warning($MSG_PASS_NOT_SET, $klms_ctrl);
                print "\n\n";
                last;
            }
            system($klms_ctrl, '--set-web-admin-password');
            $result = ($? == 0);
            print "\n";

            if ($result) {
                if (defined($self->{'AI_SAVE_FILE'})) {
                    KL::Utils::warning($MSG_WMCONSOLE_PASS_WARNING, $klms_ctrl);
                }
                last;
            }
        }
    }
}

sub set_klms_facade_settings
{
    my $self=shift;
    my $settings = shift;
    my $klms_ctrl = get_klms_ctrl();

    my $file = "/tmp/klms_settings_klmsuigenerated";
    if(open(my $fh,">", $file)) {
       print $fh $settings;
       close $fh;
       my ($result, $output) = KL::Utils::execute($klms_ctrl, '--set-settings',4,'--file', $file);
       unlink $file;
       return ($result ==0);
    }else{
        KL::Utils::error("Failed create file $file");
        return 0;
    }

}

sub enable_web_in_lms_if_installed
{
    my $self = shift;
    my $port = shift;

    my $var_root = $self->appinfo('paths','varroot');
    my $rcd = $self->appinfo('paths','rcd');
    my $klms_installer_dat = "$var_root/klms/installer.dat";

    if ( -f $klms_installer_dat ) {
        if (KL::LmsUtils::service_is_running('klms')) {
            my $settings_facade = $self->get_klms_facade_settings();
            $settings_facade =~ s/<enableWebService>(\d*)<\/enableWebService>/<enableWebService>1<\/enableWebService>/s ;
            my $already_enabled = $1;
            if ("$already_enabled" ne "1"){
                my $klms_ctrl = get_klms_ctrl();
                my ($result, $output) = KL::Utils::execute($klms_ctrl, "--start-web-service");
                if ($result !=0) {
                    KL::Utils::error("Can't start web support: $output");
                }
            }
            $settings_facade =~ s/<port>.*<\/port>/<port>$port<\/port>/s;
            return  $self->set_klms_facade_settings($settings_facade);
        }
    } 
}

sub get_apache_architecture
{
    my $httpd_path = shift;

    my ($ret, $out) = get_apache_compile_options($httpd_path);

    if ($ret != 0) {
        KL::Utils::error("Cannot get apache compile options to determine Apache architecture. ");
        return undef;
    }
    if ($out =~ /Architecture:\s+(\d+)-bit/) {
        my $apache_arch = "$1";
        KL::Utils::debug("Apache architecture detected: $apache_arch");
        return $apache_arch;
    }
    KL::Utils::error("Cannot parse output of 'apachectl -V' command to determine Apache architecture. See this output below:\n$out");
    return undef;
}

sub do_setup_apache_module
{
    my $self = shift;
    my $app_arch = $self->appinfo('product', 'arch');

    KL::Utils::header($HEADER_APACHE_MODULE);

    my $apache = $self->_ask_apache_paths();
    KL::Utils::debug("Apache configuration: " . Dumper($apache));

    if (!$apache) {
        return 0;
    }

    if ($app_arch =~ /64/ ){
        $app_arch = "64";
    }else{
        $app_arch = "32";
    }
    my $apache_arch = get_apache_architecture();

    if ( !defined($apache_arch) ) {
        my $do_continue= KL::Utils::ask_boolean("Can't determine Apache architecture. Continue with installation?");
        $self->ai_save('IGNORE_APACHE_ARCH', $do_continue?'yes':'no');
        if (! $do_continue ) {
            exit 1;
        }
    } elsif ($apache_arch ne $app_arch) {
        KL::Utils::fatal(1, "Package architecture [%1-bit] is incompatible with Apache architecture [%2-bit]. Please uninstall klmsui and install correct version.", $app_arch, $apache_arch );
    }

    $self->ai_save('APACHE_BIN',    $apache->{'bin'});
    $self->ai_save('APACHE_CONF_D', $apache->{'conf.d'});
    $self->ai_save('APACHE_INIT_D', $apache->{'service_control'});

    # ask for virtual host type
    my $vhost_type = 'port';
    my $vhost_port = '9045';
    my $vhost_host = _get_hostname('localhost');
    my $vhost_dir = 'klms';

    my $default_vhost_type = $vhost_type; 
    while (1) {
        $vhost_type = KL::Utils::ask_question( $default_vhost_type, $MSG_APACHE_ASK_VHOST_TYPE);

        last if ($vhost_type =~ /^(port|name|dir)$/);

        KL::Utils::error($MSG_APACHE_BAD_VHOST_TYPE, $vhost_type);
        print "\n";
    }

    $self->ai_save('VHOST_TYPE', $vhost_type);

    if ($vhost_type eq 'port') {
        # ask parammeters for port based template
        my $default_vhost_port = $vhost_port; 
        while (1) {
            $vhost_port = KL::Utils::ask_question( $default_vhost_port, $MSG_APACHE_ASK_VHOST_PORT);
            last if ($vhost_port =~ /^\d+$/);
            KL::Utils::error($MSG_APACHE_BAD_VHOST_PORT, $vhost_port);
            print "\n";
        }
        $self->ai_save('VHOST_PORT', $vhost_port);
    } elsif ($vhost_type eq 'name') {
        KL::Utils::warning( $MSG_NAME_VHOST_NOTICE);
        print "\n";
        # ask parameters for name based template
        while (1) {
            my $default_vhost_host = "klms.$vhost_host";
            $vhost_host = KL::Utils::ask_question( $default_vhost_host, $MSG_APACHE_ASK_VHOST_HOST);
            last if _is_valid_hostname($vhost_host);

            KL::Utils::error($MSG_APACHE_BAD_VHOST_HOST, $vhost_host);
            print "\n";
        }
        $self->ai_save('VHOST_HOST', $vhost_host);
    } elsif ($vhost_type eq 'dir') {
        KL::Utils::warning( $MSG_DIR_SECURITY_WARNING);
        print "\n";
	# ask parameters for directory-based template
        while (1) {
            my $default_vhost_dir = $vhost_dir;
            $vhost_dir = KL::Utils::ask_question( $default_vhost_dir, $MSG_APACHE_ASK_VHOST_DIR);
            last if _is_valid_url_subdir($vhost_dir);

            KL::Utils::error($MSG_APACHE_BAD_VHOST_DIR, $vhost_dir);
            print "\n";
        }
        # remove first/last slash if present
        $vhost_dir =~ s@^\/?(.*?)\/?$@$1@;
        $self->ai_save('VHOST_DIR', $vhost_dir);

    } else {
        KL::Utils::fatal(1, "Unsupported vhost type: $vhost_type");
    }

    ### ask for KL product host name
    my $ui_host = '127.0.0.1:2711';
    while (1) {
        my $default_ui_host = $ui_host;
        $ui_host = KL::Utils::ask_question( $default_ui_host, $MSG_APACHE_ASK_UI_HOST);

        last if ($ui_host =~ /^(.+?)(:\d+)$/);

        KL::Utils::error($MSG_APACHE_BAD_UI_HOST, $ui_host);
        print "\n";
    }

    $self->ai_save('UI_HOST', $ui_host);
    if ($ui_host =~ /^127.0.0.1:(\d+)$/) {
        my $port = $1;
        if ($self->enable_web_in_lms_if_installed($port) ) {
            $self->set_web_password();
        }
    }
    ### ask for certificate parameters
    my $cert_type = 'new';

    my @valid_types = qw(new file);
    my $message = $MSG_APACHE_ASK_CERT_TYPE;
    my $error = $MSG_APACHE_BAD_CERT_TYPE;

    my $name = $self->appinfo('product', 'name');
    my $cert_key = sprintf('%s/.%s.key', $apache->{'conf.d'}, $name);
    my $cert_crt = sprintf('%s/.%s.crt', $apache->{'conf.d'}, $name);

    if ( (-f $cert_key) && (-f $cert_crt) ) {
        $message = $MSG_APACHE_ASK_CERT_TYPE_W_KEEP;
        $error = $MSG_APACHE_BAD_CERT_TYPE_W_KEEP;
        push @valid_types, "keep";
        $cert_type = "keep";
    }

    # if we use directory-based config, certificate is not used in default config.
    if ($vhost_type ne 'dir')
    {
        while (1) {
            $cert_type = KL::Utils::ask_question ($cert_type, $message);

            last if grep(/^$cert_type$/, @valid_types);

            KL::Utils::error($error, $cert_type);
            $cert_type = 'new';
            print "\n";
        }
    }

    $self->ai_save('CERT_TYPE', $cert_type);

    $cert_key = undef;
    $cert_crt = undef;

    if ($cert_type eq 'new') {
        # nothing
    } elsif ($cert_type eq 'keep') {
        # nothing
    } elsif ($cert_type eq 'file') {
        while (1) {
            $cert_key = KL::Utils::ask_question ($cert_key, $MSG_APACHE_ASK_CERT_KEY);
            $cert_key = KL::Utils::detildefy($cert_key);
            last if -f $cert_key;

            KL::Utils::error($MSG_APACHE_BAD_CERT_KEY, $cert_key);
            $cert_key = undef;
            print "\n";
        }
        $self->ai_save('CERT_KEY', $cert_key);

        while (1) {
            $cert_crt = KL::Utils::ask_question ($cert_crt, $MSG_APACHE_ASK_CERT_CRT);
            $cert_crt = KL::Utils::detildefy($cert_crt);
            last if -f $cert_crt;

            KL::Utils::error($MSG_APACHE_BAD_CERT_CRT, $cert_crt);
            $cert_crt = undef;
            print "\n";
        }
        $self->ai_save('CERT_CRT', $cert_crt);
    } else {
        KL::Utils::fatal(1, "Unsupported cert type: $cert_type");
    }

    my $opts = {
        apache => $apache,

        vhost_type => $vhost_type,
        vhost_port => $vhost_port,
        vhost_host => $vhost_host,
        vhost_dir => $vhost_dir,

        ui_host  => $ui_host,

        cert_type => $cert_type,
        cert_key  => $cert_key,
        cert_crt  => $cert_crt,
    };

    $self->_setup_apache($opts);
}

sub _check_EULA_l10n
{
    my $self = shift;
    my $name = shift;
    my $filepath = shift;

    my $res = 0;
    foreach my $lang (keys %{$self->{'_LMS_L10N'}}) {
        if ($name =~ /^$lang(-.+)?$/) {
            $res = 1;
            last;
        }
    }

    KL::Utils::debug("'$name' '$filepath' = $res");

    return $res;
}


sub _check_lang_versions
{
    my $self = shift;

    my $pkgname = $self->appinfo('product', 'name');
    my $pkgver  = $self->appinfo('product', 'version');
    my $pkgtype = $self->appinfo('product', 'package');

    my $klms_share = KL::Utils::is_freebsd() ? '/usr/local/share/klms' : '/opt/kaspersky/klms/share';
    my $klms_locale = "$klms_share/locale";

    $self->{'_LMS_L10N'} = {};
    $self->{'_LMS_L10N'}->{'English'} = 'en_US';

    return if not -d $klms_locale;

    my $found = 0;

    my $dir;
    opendir ($dir, $klms_locale) || KL::Utils::fatal(1, 'Can\'t open directory %1', $klms_locale);
    my @langs = grep { -d "$klms_locale/$_" and $_ !~ /^(\.\.?|en_US)$/ } readdir($dir);
    closedir $dir;

    foreach my $l (@langs) {
        my $incompat = 0;

        if ( -f "$klms_locale/$l/locale.name" ) {
            $incompat = 1;
        } elsif ( -f "$klms_locale/$l/lang-name.txt" ) {
            $incompat = 1;
            my $verfile = "$klms_locale/$l/lang-version.txt";
            if ( -f $verfile ) {
                my $version = KL::Utils::read_file($verfile);
                $version =~ s/\n.*//g; # keep only first line
                my ($ver, $release) = split(/\s+/, $version, 2);
                if ( $ver eq $pkgver ) {
                    $incompat = 0;

                    my $info = KL::Utils::read_file("$klms_locale/$l/lang-name.txt");
                    $info =~ s/\n.*//g; # keep only first line
                    my ($name, $localized) = split(/\s+/, $info, 2);
                    $self->{'_LMS_L10N'}->{$name} = $l;
                }
            }
        }

        if ( $incompat ) {
            if ( not $found ) {
                print "Found localization package(s) incompatible with $pkgname ver. $pkgver:\n";
                $found = 1;
            }

            my ($lang, $region) = split(/_/, $l, 2);
            $lang = 'cn' if ($lang eq 'zh');

            my $lang_pkg;
            if ($pkgtype eq 'deb') {
                $lang_pkg = "klms-l10n-$lang";
            } else {
                $lang_pkg = "klms_$lang";
            }
            print "    $lang_pkg\n";
        }
    }

    if ($found) {
        KL::Utils::fatal(1, "Please uninstall it before executing %1.", $self->{'SETUP'});
    }
}

sub _app_show_EULAs
{
    my $self = shift;

    $self->_check_lang_versions();

    # show application license
    my $docs_path = $self->get_path('docs');

    my $eula_file = "$docs_path/LICENSE";
    my $agreed = $self->is_EULA_agreed('WEB_EULA_AGREED', $eula_file);
    if ($agreed ne 'yes') {
        $self->show_EULA($eula_file, 'WEB_EULA_AGREED');
    }
}

sub _update_web_eula_agreed
{
    my $self = shift;
    my $value = shift;

    my $apache_conf = $self->settings('APACHE_CONF');
    if (not defined $apache_conf or (not -f $apache_conf)) {
        return 0;
    }

    my $name = 'KlmsEulaAgreed';

    KL::Utils::debug("update to '$name $value' in apache config '$apache_conf'");

    my $conf = KL::Utils::read_file($apache_conf);
    if (not defined $conf) {
        KL::Utils::debug("failed read '$apache_conf': $!");
        return 0;
    }

    my $new_conf = '';
    my $found = 0;

    my @lines = split(/\n/, $conf);
    my $cnt = 0;
    foreach my $line (@lines) {
        $cnt++;

        if ($line =~ /^\s*(FastCgiExternalServer\s|<.+).*$/) {
            # a tag or FastCgiExternalServer command found
            if (not $found) {
                KL::Utils::debug("+line $cnt added ($1)");
                $line = "$name $value\n\n$line";
                $found = 2;
            }
        } elsif ($line =~ /^\s*$name\s+(.*?)\s*$/) {
            if ($found) {
                KL::Utils::debug("-line $cnt commented out");
                $line = "# $line";
            } else {
                KL::Utils::debug("*line $cnt updated (was '$1')");
                $line = "$name $value";
            }
            $found = 1;
        }

        $new_conf .= "$line\n";
    }

    my $res = KL::Utils::write_file($apache_conf, $new_conf);
    if (not $res) {
        KL::Utils::fatal(1, "Cannot write to file '%1': <BR>%2", $apache_conf, $!);
    }

    KL::Utils::debug("success");

    return 1;
}

sub settings
{
    my $self = shift;

    my $name = $_[0];
    my $value = $_[1];

    if ((defined $value) and ($name eq 'WEB_EULA_AGREED')) {
        $self->_update_web_eula_agreed($value);
    }

    return $self->SUPER::settings(@_);
}

sub remove_settings
{
    my $self = shift;

    foreach my $name (@_) {
        if ($name eq 'WEB_EULA_AGREED') {
            $self->_update_web_eula_agreed('no');
        }
    }

    return $self->SUPER::settings(@_);
}

sub do_postupgrade
{
    my $self = shift;

    $self->_app_show_EULAs();

    my $upgrade_type = $self->settings('UPGRADE_TYPE');

    my $configured = 0;

    if (defined $upgrade_type and ($upgrade_type eq 'std')) {
        my $conf = $self->settings('APACHE_CONF');
        my $service_name = $self->settings('APACHE_INIT_D');

        if (defined($conf) and -f $conf and defined($service_name)) {
            $configured = 1;

            # implement try-restart manually (it is not supported by systemd)
            if (KL::LmsUtils::service_is_running("$service_name")) {
                KL::LmsUtils::service_restart( $service_name);
            }
        }
    }

    if (not $configured) {
        $self->do_setup_apache_module();
    }

    $self->remove_settings('UPGRADE_TYPE');

    KL::Utils::output('Upgrade completed.');
}

sub do_postinstall
{
    my $self = shift;

    $self->_app_show_EULAs();

    $self->do_setup_apache_module();

    return 1;
}

sub do_auto_install
{
    my $self = shift;

    my ($opt, $res, $message);

    my $vhost_type = 'port';

    my $default_vhost_port = '9045';
    my $vhost_port = $default_vhost_port;

    my $default_vhost_host = _get_hostname('localhost');
    my $vhost_host = $default_vhost_host;

    my $default_vhost_dir = 'klms';
    my $vhost_dir = $default_vhost_dir;

    # check EULA agreement
    $opt = $self->ai_value('WEB_EULA_AGREED');
    if ($opt ne 'yes') {
        KL::Utils::error('EULA is not agreed');
        return 0;
    }

    my $docs_path = $self->get_path('docs');
    my $eula_file = "$docs_path/LICENSE";
    $self->settings('WEB_EULA_AGREED', 'yes'); # not WEB_EULA because installer.dat differs for klms and klmsui
    $self->settings('WEB_EULA_AGREED_VER', $self->_md5sum($eula_file));
    $self->ai_save('WEB_EULA_AGREED','yes');

    my $app_arch = $self->appinfo('product', 'arch');
    KL::Utils::header($HEADER_APACHE_MODULE);

    my $apache;
    $apache->{'bin'} = $self->ai_value('APACHE_BIN'); 
    $apache->{'conf.d'} = $self->ai_value('APACHE_CONF_D'); 
    $apache->{'service_control'} = $self->ai_value('APACHE_INIT_D');

    if ($app_arch =~ /64/ ){
        $app_arch = "64";
    } else {
        $app_arch = "32";
    }
    my $apache_arch = get_apache_architecture();

    if (!defined($apache_arch)) {
        my $yes_no = $self->ai_value('IGNORE_APACHE_ARCH');
        if ( !defined($yes_no) || ( $yes_no =~ /^no$/i ) ) { 
            KL::Utils::error("Can't determine Apache architecture.");
            $self->_ai_error(1);
        }
    } elsif ($apache_arch ne $app_arch) {
        KL::Utils::error("Package architecture [%1-bit] is incompatible with Apache architecture [%2-bit]. Please uninstall klmsui and install correct version.", $app_arch, $apache_arch );
        $self->_ai_error(1);
    }

    $self->ai_save('APACHE_CONF_D', $apache->{'conf.d'});
    $self->ai_save('APACHE_INIT_D', $apache->{'service_control'});

    # ask for virtual host type
    $vhost_type = $self->ai_value('VHOST_TYPE');

    $self->ai_save('VHOST_TYPE', $vhost_type);

    if ($vhost_type eq 'port') {
            $vhost_port = $self->ai_value('VHOST_PORT');
            if (KL::LmsUtils::is_empty( $vhost_port))
            {
                $vhost_port = $default_vhost_port;
                KL::Utils::warning("Parameter 'VHOST_PORT' is not set in auto-install file, using default: \'$vhost_port\'");
            }
            $self->ai_save('VHOST_PORT', $vhost_port);
    } elsif ($vhost_type eq 'name') {
            $vhost_host = $self->ai_value('VHOST_HOST');
            if (KL::LmsUtils::is_empty( $vhost_host))
            {
                $vhost_host = $default_vhost_host;
                KL::Utils::warning("Parameter 'VHOST_HOST' is not set in auto-install file, using default: \'$vhost_host\'");
            }
            $self->ai_save('VHOST_HOST', $vhost_host);
    } elsif ($vhost_type eq 'dir') {
            $vhost_dir = $self->ai_value('VHOST_DIR');
            if (KL::LmsUtils::is_empty( $vhost_dir)) {
                $vhost_dir = $default_vhost_dir;
                KL::Utils::warning("Parameter 'VHOST_DIR' is not set in auto-install file, using default: \'$vhost_dir\'");
            }
            $self->ai_save('VHOST_DIR', $vhost_dir);
    } else {
        KL::Utils::error("Unsupported vhost type: $vhost_type");
        $self->_ai_error(2);
    }

    ### KL product host name
    my $ui_host = $self->ai_value('UI_HOST');
    if ($ui_host !~ /^(.+?)(:\d+)$/) {
        KL::Utils::error($MSG_APACHE_BAD_UI_HOST, $ui_host);
        $self->_ai_error(2);
    }
    $self->ai_save('UI_HOST', $ui_host);

    ### certificate parameters
    my $cert_type = $self->ai_value('CERT_TYPE');
    if ($cert_type !~ /^new|file|keep$/){
        KL::Utils::error($MSG_APACHE_BAD_CERT_TYPE, $cert_type);
        $self->_ai_error(2);
    }
    $self->ai_save('CERT_TYPE', $cert_type);

    my ($cert_key, $cert_crt);
    if ($cert_type eq 'new') {
        # nothing
    } elsif ($cert_type eq 'keep') {
        my $name = $self->appinfo('product', 'name');
        my $key = sprintf('%s/.%s.key', $apache->{'conf.d'}, $name);
        my $crt = sprintf('%s/.%s.crt', $apache->{'conf.d'}, $name);

        if (! -f $key) {
            KL::Utils::error("Cannot keep existing certificate, '%1' file not found", $key);
            $self->_ai_error(3);
        }
        if (! -f $crt) {
            KL::Utils::error("Cannot keep existing certificate, '%1' file not found", $crt);
            $self->_ai_error(3);
        }
    } elsif ($cert_type eq 'file') {
        $cert_key = $self->ai_value('CERT_KEY');
        if (! -f $cert_key ){
            KL::Utils::error($MSG_APACHE_BAD_CERT_KEY, $cert_key);
            $self->_ai_error(2);
        }
        $self->ai_save('CERT_KEY', $cert_key);
        $cert_crt = $self->ai_value('CERT_CRT');
        if (!-f $cert_crt) {
            KL::Utils::error($MSG_APACHE_BAD_CERT_CRT, $cert_crt);
            $self->_ai_error(2);
        }
        $self->ai_save('CERT_CRT', $cert_crt);
    } else {
        KL::Utils::error("Unsupported cert type: $cert_type");
        $self->_ai_error(2);
    }

    my $opts = {
        apache => $apache,

        vhost_type => $vhost_type,
        vhost_port => $vhost_port,
        vhost_host => $vhost_host,
        vhost_dir => $vhost_dir,

        ui_host  => $ui_host,

        cert_type => $cert_type,
        cert_key  => $cert_key,
        cert_crt  => $cert_crt,
    };

    $self->_setup_apache($opts);

    $self->save_settings();

    return 1;
}

sub _breaked_upgrade_detected
{
    my $self = shift;

    my $value;

    $value = $self->settings('WEB_EULA_AGREED');
    if (not defined $value or $value ne 'yes') {
        return 0;
    }

    KL::Utils::debug('YES');

    return 1;
}

sub trigger_install
{
    my $self = shift;
    my $values = shift;

    my $upgrade = defined $values->{'upgrade'};
    my $upgrade_type;

    if ($upgrade) {
        $upgrade_type = 'std';
    } elsif ($self->_breaked_upgrade_detected()) {
        $upgrade = 1;
        $values->{'upgrade'} = 1;
        $upgrade_type = 'brk';
    }

    if (defined $upgrade_type) {
        $self->settings('UPGRADE_TYPE', $upgrade_type);
        $self->save_settings();
    }

    return $self->SUPER::trigger_install($values);
}

sub do_uninstall
{
    my $self = shift;
    my $upgrade = shift;

    return 1 if $upgrade;

    my $conf = $self->settings('APACHE_CONF');

    # if klmsui-setup was not completed or not executed at all, $conf is empty.
    if (KL::LmsUtils::is_empty( $conf) || ! -f $conf) {
        return 1;
    }

    KL::Utils::header($HEADER_UPDATE_APACHE);

    my $conf_bak = $conf.'.backup~';

    unlink $conf_bak if -f $conf_bak;
    if (!rename($conf, $conf_bak)) {
         KL::Utils::warning( $MSG_CANT_WRITE, $conf_bak, $!);
    } else {
        KL::Utils::output( $MSG_CONFIG_SAVED, $conf_bak);
    }

    my $service_control = $self->settings('APACHE_INIT_D');
    if (defined($service_control) && KL::LmsUtils::find_service( $service_control)) {
        print "\n";
        KL::LmsUtils::service( $service_control, $self->{'APACHE_RELOAD_CMD'});
    }

    return 1;
}

sub ai_validate_service
{
    my $self = shift;
    my $rule = shift;
    my $service_name = shift;
    if (KL::LmsUtils::find_service( $service_name)) {
        return 1;
    }
    return (0, "not a valid service name or rc.d script name");
}

sub ai_validate_port
{
    my $self = shift;
    my $rule = shift;
    my $port = shift;
    return (is_valid_port($port));
}

sub is_valid_port
{
    my $port = shift;
    if ($port=~ /^\d+$/) {
        if ($port >0 && $port <65535){
            return 1;
        }
    }
    return 0;
}

sub ai_validate_hostandport
{
    my $self = shift;
    my $rule = shift;
    my $hostandport = shift;
    if ($hostandport =~ /^(.*):(\d+)/ ) {
        my ($ip,$port) =($1,$2);
        if (is_valid_port($port) && (_is_valid_hostname ($ip) || _is_valid_ip_address($ip) )) {
            return 1;
        }
    }
    return 0;
}

sub ai_validate_url_subdir
{
    my $self = shift;
    my $rule = shift;
    my $url_subdir = shift;
    return (_is_valid_url_subdir($url_subdir));
}

1;
